/*
** File: vxbus_pci.c  
** Project: VxWorks Driver Framework
** Purpose: VxBus entry points
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

#if defined(ADB3_VXBUS)

# include <hwif/vxbus/vxbPciLib.h>

static void
dfVxbInstInit(
  VXB_DEVICE_ID pInst);

static void
dfVxbInstInit2(
  VXB_DEVICE_ID pInst);

static void
dfVxbInstConnect(
  VXB_DEVICE_ID pInst);

static BOOL
dfVxbDevProbe(
  struct vxbDev* pVxbDev);

static STATUS
dfVxbDrvUnlink(
  VXB_DEVICE_ID pInst);

/*
** ----------------------------------------------------------------------
** Define vxBus-related data structures
** ----------------------------------------------------------------------
*/

static DfDriverObject g_dfDrvObj;
static boolean_t g_bPciDriverRegistered = FALSE;

static DfSpinLock g_pciConfigLock;

static struct drvBusFuncs g_dfVxbFuncs = {
  dfVxbInstInit,     /* devInstanceInit */
  dfVxbInstInit2,    /* devInstanceInit2 */
  dfVxbInstConnect   /* devConnect */
};

static device_method_t g_dfVxbMethods [] = {
  DEVMETHOD(vxbDrvUnlink, dfVxbDrvUnlink),
  { 0, 0 } /* Terminator */
};

static struct vxbPciID* g_pDfVxbPciIdList = NULL;
static DfPciCompatibleId* g_pPciCompatibleIdTable = NULL;
static unsigned int g_numPciCompatibleId = 0;

static struct vxbPciRegister g_dfVxbDevRegistration = {
  {
    NULL,               /* pNext */
    VXB_DEVID_DEVICE,   /* devID */
    VXB_BUSID_PCI,      /* busID = PCI */
#if defined(DFVXW_VXBUS5)
    VXB_VER_5_0_0,      /* vxbVersion */
#elif defined(DFVXW_VXBUS4)
    VXB_VER_4_0_0,      /* vxbVersion */
#elif defined(DFVXW_VXBUS3)
    VXBUS_VERSION_3,    /* vxbVersion */
#else
# error None of DFVXW_VXBUS{3,4,5} defined!
#endif
    "?",                /* drvName */
    &g_dfVxbFuncs,      /* pDrvBusFuncs */
    g_dfVxbMethods,     /* pMethods */
    dfVxbDevProbe,      /* devProbe */
    NULL                /* pParamDefaults */
  },                    /* b */
  0,                    /* initialized by registerDriver! */
  NULL                  /* initialized by registerDriver! */
};

/*
** ----------------------------------------------------------------------
** Helper functions
** ----------------------------------------------------------------------
*/

#if defined(DFVXW_VXBUS3)

# define pciConfigRead8(pDev, addr, pVal8, pFlags) pDev->pAccess->busCfgRead(pDev, addr, sizeof(*pVal8), (char*)pVal8, pFlags)
# define pciConfigRead16(pDev, addr, pVal16, pFlags) pDev->pAccess->busCfgRead(pDev, addr, sizeof(*pVal16), (char*)pVal16, pFlags)
# define pciConfigRead32(pDev, addr, pVal32, pFlags) pDev->pAccess->busCfgRead(pDev, addr, sizeof(*pVal32), (char*)pVal32, pFlags)
# define pciConfigWrite8(pDev, addr, pVal8, pFlags) pDev->pAccess->busCfgWrite(pDev, addr, sizeof(*pVal8), (char*)pVal8, pFlags)
# define pciConfigWrite16p(pDev, addr, pVal16, pFlags) pDev->pAccess->busCfgWrite(pDev, addr, sizeof(*pVal16), (char*)pVal16, pFlags)
# define pciConfigWrite32(pDev, addr, pVal32, pFlags) pDev->pAccess->busCfgWrite(pDev, addr, sizeof(*pVal32), (char*)pVal32, pFlags)

#elif defined(DFVXW_VXBUS4) || defined(DFVXW_VXBUS5)

# define pciConfigRead8(pDev, addr, pVal8, pFlags) vxbPciDevCfgRead(pDev, addr, sizeof(*pVal8), pVal8)
# define pciConfigRead16(pDev, addr, pVal16, pFlags) vxbPciDevCfgRead(pDev, addr, sizeof(*pVal16), pVal16)
# define pciConfigRead32(pDev, addr, pVal32, pFlags) vxbPciDevCfgRead(pDev, addr, sizeof(*pVal32), pVal32)
# define pciConfigWrite8(pDev, addr, pVal8, pFlags) vxbPciDevCfgWrite(pDev, addr, sizeof(*pVal8), *(pVal8))
# define pciConfigWrite16(pDev, addr, pVal16, pFlags) vxbPciDevCfgWrite(pDev, addr, sizeof(*pVal16), *(pVal16))
# define pciConfigWrite32(pDev, addr, pVal32, pFlags) vxbPciDevCfgWrite(pDev, addr, sizeof(*pVal32), *(pVal32))

#else

# error Neither DFVXW_VXBUS3 nor DFVXW_VXBUS{4,5} defined!

#endif

/* Read the BARs of the PCI device (raw resources) */
static void
getRawBusResources(
  DfBusResources* pResources,
  VXB_DEVICE_ID pVxbDev)
{
  static const UINT32 barAddresses[] = {
    PCI_CFG_BASE_ADDRESS_0,
    PCI_CFG_BASE_ADDRESS_1,
    PCI_CFG_BASE_ADDRESS_2,
    PCI_CFG_BASE_ADDRESS_3,
    PCI_CFG_BASE_ADDRESS_4,
    PCI_CFG_BASE_ADDRESS_5
  };
  const uint32_t allOnes = ~(uint32_t)0;
  unsigned int i;
  unsigned int ioBarCount, memBarCount;
  boolean_t b64Bit, bIo;
  uint32_t currLo, currHi;
  uint32_t maskLo, maskHi;
  uint8_t flags;
  uint8_t intLine, intPin;
  uint64_t mask64;
  UINT32 f = 0;

  /* Get the BAR sizes */
  ioBarCount = 0;
  memBarCount = 0;
  for (i = 0; i < DF_ARRAY_LENGTH(barAddresses); i++) {
    flags = 0;
    pciConfigRead32(pVxbDev, barAddresses[i], &currLo, &f);
    if (currLo & PCI_BASE_IO) {
      currLo &= PCI_IOBASE_MASK;
      b64Bit = FALSE;
      bIo = TRUE;
    } else {
      if (currLo & PCI_BASE_IN_64BITS) {
        b64Bit = TRUE;
      } else {
        b64Bit = FALSE;
      }
      if (currLo & PCI_BASE_PREFETCH) {
        flags |= DF_PCIBARF_PREFETCH;
      }
      flags = (uint8_t)(currLo & ~PCI_MEMBASE_MASK);
      flags = (uint8_t)(flags & 0xFEU);
      currLo &= PCI_MEMBASE_MASK;
      bIo = FALSE;
    }
    if (b64Bit) {
      /* 64-bit BAR */
      if (i == 5) {
        continue;
      }
      pciConfigRead32(pVxbDev, barAddresses[i + 1], &currHi, &f);
      pciConfigWrite32(pVxbDev, barAddresses[i], &allOnes, &f);
      pciConfigWrite32(pVxbDev, barAddresses[i + 1], &allOnes, &f);
      pciConfigRead32(pVxbDev, barAddresses[i], &maskLo, &f);
      pciConfigRead32(pVxbDev, barAddresses[i + 1], &maskHi, &f);
      maskLo &= PCI_MEMBASE_MASK;
      pciConfigWrite32(pVxbDev, barAddresses[i], &currLo, &f);
      pciConfigWrite32(pVxbDev, barAddresses[i + 1], &currHi, &f);
      if ((maskLo == 0 && maskHi == 0) || (maskLo != 0 && (b64Bit && maskHi != 0xffffffffUL))) {
          /* BAR size is 0 (or invalid) */
      } else {
        mask64 = (uint64_t)maskLo;
        mask64 |= ((uint64_t)maskHi) << 32;
        dfAssert(mask64 != 0xffffffffffffffffULL);
        pResources->variant.pci.memBar[memBarCount].size = ~mask64 + 1;
        pResources->variant.pci.memBar[memBarCount].base = (uint64_t)currLo;
        pResources->variant.pci.memBar[memBarCount].base |= ((uint64_t)currHi) << 32;
        pResources->variant.pci.memBar[memBarCount].flags = flags;
        memBarCount++;
      }
      i++;
    } else {
      /* 32-bit BAR */
      pciConfigWrite32(pVxbDev, barAddresses[i], &allOnes, &f);
      pciConfigRead32(pVxbDev, barAddresses[i], &maskLo, &f);
      maskLo &= bIo ? PCI_IOBASE_MASK : PCI_MEMBASE_MASK;
      pciConfigWrite32(pVxbDev, barAddresses[i], &currLo, &f);
      if (maskLo == 0) {
        /* BAR size is 0 */
      } else {
        dfAssert(maskLo != 0xffffffffU);
        if (bIo) {
          pResources->variant.pci.ioBar[ioBarCount].base = (uint64_t)currLo;
          pResources->variant.pci.ioBar[ioBarCount].size = (uint64_t)(~maskLo + 1);
          pResources->variant.pci.ioBar[ioBarCount].flags = flags;
          ioBarCount++;
        } else {
          pResources->variant.pci.memBar[memBarCount].base = (uint64_t)currLo;
          pResources->variant.pci.memBar[memBarCount].size = (uint64_t)(~maskLo + 1);
          pResources->variant.pci.memBar[memBarCount].flags = flags;
          memBarCount++;
        }
      }
    }
  }
  pResources->variant.pci.numMemBar = memBarCount;
  pResources->variant.pci.numIoBar = ioBarCount;

  pciConfigRead8(pVxbDev, PCI_CFG_DEV_INT_PIN, &intPin, &f);
  if (intPin == 0) {
    pResources->variant.pci.numIrq = 0;
  } else {
    pciConfigRead8(pVxbDev, PCI_CFG_DEV_INT_LINE, &intLine, &f);
    pResources->variant.pci.numIrq = 1;
    pResources->variant.pci.irq[0].bMessageSignalled = FALSE;
    pResources->variant.pci.irq[0].bLevelSensitive = TRUE;
    pResources->variant.pci.irq[0].variant.raw.intPin = intPin;
    pResources->variant.pci.irq[0].variant.raw.intLine = intLine;
  }
}

static STATUS
registerDriver(
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  STATUS err = OK;
  DfPciCompatibleId* pPciCompatibleId;
  unsigned int i, j, nId, nLegacyId, nEntry;
  BOOL bUnique;

  /* Count the number of PCI hardware IDs */
  for (nId = 0; pPciCompatibleIdTable[nId].match != 0xffU; nId++);
  g_numPciCompatibleId = nId;

  if (bLegacyHardware) {
    /* Count the number of legacy PCI hardware IDs */
    for (nLegacyId = 0; pLegacyPciCompatibleIdTable[nLegacyId].match != 0xffU; nLegacyId++);
    g_numPciCompatibleId += nLegacyId;
  } else {
    nLegacyId = 0;
  }

  /* Make our own compatible PCI ID table in case the ones passed in become invalid later */
  g_pPciCompatibleIdTable = (DfPciCompatibleId*)hwMemAlloc(sizeof(DfPciCompatibleId) * g_numPciCompatibleId);
  if (NULL == g_pPciCompatibleIdTable) {
    dfDebugPrint(0, ("registerDriver: failed to allocate PCI Compatible ID table\n"));
    err = ERROR;
    goto done;
  }
  bcopy((const char*)pPciCompatibleIdTable, (char*)g_pPciCompatibleIdTable, sizeof(DfPciCompatibleId) * nId);
  if (bLegacyHardware) {
    bcopy((const char*)pLegacyPciCompatibleIdTable, (char*)(g_pPciCompatibleIdTable +  nId), sizeof(DfPciCompatibleId) * nLegacyId);
  }

  g_pDfVxbPciIdList = (struct vxbPciID*)hwMemAlloc(sizeof(struct vxbPciID) * g_numPciCompatibleId);
  if (NULL == g_pDfVxbPciIdList) {
    dfDebugPrint(0, ("registerDriver: failed to allocate PCI ID table\n"));
    err = ERROR;
    goto done;
  }

  /*
  ** Enumerate the unique PCI { vendor, device } pairs; can be less than the
  ** number of hardware IDs in the driver object.
  */
  nEntry = 0;
  for (i = 0; i < g_numPciCompatibleId; i++) {
    pPciCompatibleId = &g_pPciCompatibleIdTable[i];
    bUnique = TRUE;
    for (j = 0; j < nEntry; j++) {
      if (pPciCompatibleId->id.device == g_pDfVxbPciIdList[j].pciDevId &&
          pPciCompatibleId->id.vendor == g_pDfVxbPciIdList[j].pciVendId) {
        bUnique = FALSE;
        break;
      }
    }

    if (bUnique) {
      g_pDfVxbPciIdList[nEntry].pciDevId = pPciCompatibleId->id.device;
      g_pDfVxbPciIdList[nEntry].pciVendId = pPciCompatibleId->id.vendor;
      nEntry++;
    }
  }

  dfDebugPrint(2, ("registerDriver: %lu unique { vendor ID, device ID } pair(s):\n", (unsigned long)nEntry));
  for (i = 0; i < nEntry; i++) {
    dfDebugPrint(2, ("registerDriver:   ID %lu { 0x%04lx, 0x%04lx }\n",
      (unsigned long)i, (unsigned long)g_pDfVxbPciIdList[i].pciVendId, (unsigned long)g_pDfVxbPciIdList[i].pciDevId));
  }
  g_dfVxbDevRegistration.idListLen = nEntry;
  g_dfVxbDevRegistration.idList = g_pDfVxbPciIdList;
  err = vxbDevRegister(&g_dfVxbDevRegistration.b);
  if (OK != err) {
    dfDebugPrint(0, ("registerDriver: failed to register driver, error=%u(0x%x)\n", (unsigned int)err, (unsigned int)err));
    goto done;
  }

done:
  if (OK != err) {
    if (NULL != g_pDfVxbPciIdList) {
      hwMemFree((char*)g_pDfVxbPciIdList);
      g_pDfVxbPciIdList = NULL;
    }
    if (NULL != g_pPciCompatibleIdTable) {
      hwMemFree((char*)g_pPciCompatibleIdTable);
      g_pPciCompatibleIdTable = NULL;
    }
  }
  return err;
}

/*
** ----------------------------------------------------------------------
** VxBus-related standard functions
** ----------------------------------------------------------------------
*/

static STATUS
dfVxbDrvUnlink(
  VXB_DEVICE_ID pVxbDev)
{
  DfDeviceObject* pDevObj;

  dfDebugPrint(1, ("dfVxbDrvUnlink: entered, pDev=%p\n", pVxbDev));

  pDevObj = (DfDeviceObject*)pVxbDev->pDrvCtrl;
  if (NULL != pDevObj) {
    dfDebugPrint(0, ("dfVxbDrvUnlink: uninitializing instance, pDevObj=%p\n", pDevObj));

    if (NULL != g_dfDrvObj.methods.pOnStop) {
      dfDebugPrint(1, ("removeDevice: calling OnStop method\n"));
      g_dfDrvObj.methods.pOnStop(pDevObj, TRUE);
    }

    if (NULL != g_dfDrvObj.methods.pOnRemove) {
      dfDebugPrint(1, ("removeDevice: calling OnRemove method\n"));
      g_dfDrvObj.methods.pOnRemove(pDevObj);
    }

    if (pDevObj->system.request.bPoolNonBlockValid) {
      dfPoolUninit(&pDevObj->system.request.poolNonBlock);	
      pDevObj->system.request.bPoolNonBlockValid = FALSE;
    }
    if (pDevObj->system.request.bPoolValid) {
      dfPoolUninit(&pDevObj->system.request.pool);	
      pDevObj->system.request.bPoolValid = FALSE;
    }

    if (NULL != pDevObj->system.pContext && g_dfDrvObj.deviceContextSize != 0) {
      dfFree(pDevObj->system.pContext);
    }
    dfFree(pDevObj);
  }

  return OK;
}

static void
dfVxbInstInit(
 VXB_DEVICE_ID pVxbDev)
{
  dfDebugPrint(1, ("dfVxbInstInit: entered, pVxbDev=%p\n", pVxbDev));

  pVxbDev->pDrvCtrl = NULL;
}

static void
dfVxbInstInit2(
 VXB_DEVICE_ID pVxbDev)
{
  DfHardwareId hardwareId;
  DfBusResources rawResources;
  DfDeviceObject* pDevObj = NULL;
  uint16_t vendor, device, subVendor, subDevice;
  uint8_t revision;
  boolean_t bFailed = FALSE, bStarted = FALSE, bAdded = FALSE;
  UINT32 f = 0;

  dfDebugPrint(1, ("dfVxbInstInit2: entered, pDev=%p\n", pVxbDev));

  pciConfigRead16(pVxbDev, PCI_CFG_VENDOR_ID, &vendor, &f);
  pciConfigRead16(pVxbDev, PCI_CFG_DEVICE_ID, &device, &f);
  pciConfigRead16(pVxbDev, PCI_CFG_SUB_VENDER_ID, &subVendor, &f);
  pciConfigRead16(pVxbDev, PCI_CFG_SUB_SYSTEM_ID, &subDevice, &f);
  pciConfigRead8(pVxbDev, PCI_CFG_REVISION, &revision, &f);

  hardwareId.busType = DfBusPci;
  hardwareId.variant.pci.vendor = vendor;
  hardwareId.variant.pci.device = device;
  hardwareId.variant.pci.subVendor = subVendor;
  hardwareId.variant.pci.subDevice = subDevice;
  hardwareId.variant.pci.revision = revision;
  getRawBusResources(&rawResources, pVxbDev);

  pDevObj = (DfDeviceObject*)dfMalloc(sizeof(*pDevObj));
  if (pDevObj == NULL) {
    dfDebugPrint(0, ("*** dfVxbInstInit2: Cannot allocate device object\n"));
    goto done;
  }
  pDevObj->system.pVxbDev = pVxbDev;
  pDevObj->system.pContext = NULL;
  pDevObj->system.pExportedInterfaces = NULL;
  pDevObj->system.pImportedInterfaces = NULL;
  pDevObj->system.pMmappableRegions = NULL;
  pDevObj->system.interfaceList.pHead = NULL;
  pDevObj->system.dma.pMapArray = NULL;
  pDevObj->system.request.bPoolValid = FALSE;
  pDevObj->system.request.bPoolNonBlockValid = FALSE;
  pDevObj->hardwareId = hardwareId;

  if (!dfPoolInit(&pDevObj->system.request.pool, pDevObj, "clReq", DfClientRequest)) {
    bFailed = TRUE;
    dfDebugPrint(0, ("*** dfVxbInstInit2: Cannot init I/O request pool\n"));
    goto done;
  }
  pDevObj->system.request.bPoolValid = TRUE;

  if (!dfPoolInit(&pDevObj->system.request.poolNonBlock, pDevObj, "clReqNB", DfVxwClientRequestNonBlock)) {
    bFailed = TRUE;
    dfDebugPrint(0, ("*** dfVxbInstInit2: Cannot init non-blocking I/O request pool\n"));
    goto done;
  }
  pDevObj->system.request.bPoolNonBlockValid = TRUE;

  if (g_dfDrvObj.deviceContextSize != 0) {
    pDevObj->system.pContext = dfMalloc(g_dfDrvObj.deviceContextSize);
    if (NULL == pDevObj->system.pContext) {
      bFailed = TRUE;
      dfDebugPrint(0, ("*** dfVxbInstInit2: Cannot allocate device context, size=%lu(0x%lx) bytes\n",
        (unsigned long)g_dfDrvObj.deviceContextSize, (unsigned long)g_dfDrvObj.deviceContextSize));
      goto done;
    }
  }

  if (NULL != g_dfDrvObj.methods.pOnAdd) {
    dfDebugPrint(1, ("dfVxbInstInit2: calling OnAdd method\n"));
    if (!g_dfDrvObj.methods.pOnAdd(pDevObj)) {
      bFailed = TRUE;
      dfDebugPrint(0, ("*** dfVxbInstInit2: OnAdd failed\n"));
      goto done;
    }
    bAdded = TRUE;
  }

  if (NULL != g_dfDrvObj.methods.pOnStart) {
    dfDebugPrint(1, ("dfVxbInstInit2: calling OnStart method\n"));
    if (!g_dfDrvObj.methods.pOnStart(pDevObj, &rawResources, &rawResources)) { /* raw and xlat resources are the same */
      bFailed = TRUE;
      dfDebugPrint(0, ("*** dfVxbInstInit2: OnStart failed\n"));
      goto done;
    }
    bStarted = TRUE;
  }

done:
  if (bFailed) {
    if (NULL != pDevObj) {
      if (bStarted) {
        g_dfDrvObj.methods.pOnStop(pDevObj, TRUE);
        bStarted = FALSE;
      }
      if (bAdded) {
        g_dfDrvObj.methods.pOnRemove(pDevObj);
        bAdded = FALSE;
      }
      if (pDevObj->system.request.bPoolNonBlockValid) {
        dfPoolUninit(&pDevObj->system.request.poolNonBlock);	
        pDevObj->system.request.bPoolNonBlockValid = FALSE;
      }
      if (pDevObj->system.request.bPoolValid) {
        dfPoolUninit(&pDevObj->system.request.pool);	
        pDevObj->system.request.bPoolValid = FALSE;
      }
      if (NULL != pDevObj->system.pContext) {
        dfFree(pDevObj->system.pContext);
        pDevObj->system.pContext = NULL;
      }
      dfFree(pDevObj);
      pDevObj = NULL;
    }
  } else {
    pVxbDev->pDrvCtrl = (void*)pDevObj;
  }
}

static void
dfVxbInstConnect(
  VXB_DEVICE_ID pVxbDev)
{
  dfDebugPrint(1, ("dfVxbInstConnect: entered, pVxbDev=%p\n", pVxbDev));
}

static BOOL
dfVxbDevProbe(
  struct vxbDev* pVxbDev)
{
  unsigned int i;
  BOOL bMatch = FALSE;
  DfPciCompatibleId* pPciCompatibleId;
  uint16_t vid, did, svid, sdid;
  uint8_t revId;
  UINT32 flags = 0;

  dfDebugPrint(1, ("dfVxbDevProbe: entered, pVxbDev=%p\n", pVxbDev));

  pciConfigRead16(pVxbDev, PCI_CFG_VENDOR_ID, &vid, &flags);
  pciConfigRead16(pVxbDev, PCI_CFG_DEVICE_ID, &did, &flags);
  pciConfigRead16(pVxbDev, PCI_CFG_SUB_VENDER_ID, &svid, &flags);
  pciConfigRead16(pVxbDev, PCI_CFG_SUB_SYSTEM_ID, &sdid, &flags);
  pciConfigRead8(pVxbDev, PCI_CFG_REVISION, &revId, &flags);
  dfDebugPrint(1, ("dfVxbDevProbe: vid=0x%04x did=0x%04x svid=0x%04x sdid=0x%04x rev=0x%02x\n",
    (unsigned int)vid, (unsigned int)did, (unsigned int)svid, (unsigned int)sdid, (unsigned int)revId));

  /* Verify that the PCI device's subsystem ID & rev ID etc. matches what we're looking for */
  for (i = 0; i < g_numPciCompatibleId; i++) {
    pPciCompatibleId = &g_pPciCompatibleIdTable[i];
    bMatch = (!(pPciCompatibleId->match & DF_PCI_MATCH_VENDOR) || vid == pPciCompatibleId->id.vendor) &&
             (!(pPciCompatibleId->match & DF_PCI_MATCH_DEVICE) || did == pPciCompatibleId->id.device) &&
             (!(pPciCompatibleId->match & DF_PCI_MATCH_SUBVENDOR) || svid == pPciCompatibleId->id.subVendor) &&
             (!(pPciCompatibleId->match & DF_PCI_MATCH_SUBDEVICE) || sdid == pPciCompatibleId->id.subDevice) &&
             (!(pPciCompatibleId->match & DF_PCI_MATCH_REVISION) || revId == pPciCompatibleId->id.revision);
    if (bMatch) {
      dfDebugPrint(1, ("dfVxbDevProbe: hardware ID %lu matches\n", (unsigned long)i));
      break;
    }
  }

  return bMatch;
}

/*
** ----------------------------------------------------------------------
** Functions for starting and stopping the driver
** ----------------------------------------------------------------------
*/

STATUS
dfVxbEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  STATUS err = OK;

#if DF_DBG_BUILD
  g_dfDebugLevel = debugLevel;
#endif
  dfVxwLogInit();

  if (!dfVxwSemPoolInit()) {
    dfDebugPrint(0, ("*** dfVxbEntry: failed to initialize binary semaphore pool\n"));
    dfVxwLogUninit();
    return ERROR;
  }

  dfDebugPrint(1, ("dfVxbEntry: entered\n"));

  dfVxwCheckDataTypes();

  dfSpinLockInit(&g_pciConfigLock);

  DF_ZERO_OBJECT(&g_dfDrvObj);
  g_dfDrvObj.methods.pOnAdd = NULL;
  g_dfDrvObj.methods.pOnRemove = NULL;
  g_dfDrvObj.methods.pOnStart = NULL;
  g_dfDrvObj.methods.pOnStop = NULL;
  g_dfDrvObj.methods.pOnPowerQuery = NULL;
  g_dfDrvObj.methods.pOnPowerSet = NULL;

  dfDebugPrint(1, ("dfVxbEntry: calling dfDriverEntry\n"));
  if (!dfDriverEntry(&g_dfDrvObj)) {
    dfVxwSemPoolUninit();
    dfVxwLogUninit();
    return ERROR;
  }

  dfDebugPrint(1, ("dfVxbEntry: registering VxBus driver\n"));
  dfStrCpy(g_dfVxbDevRegistration.b.drvName, DF_ARRAY_LENGTH(g_dfVxbDevRegistration.b.drvName), pDriverName);
  err = registerDriver(pPciCompatibleIdTable, pLegacyPciCompatibleIdTable, bLegacyHardware);
  if (OK != err) {
    dfDriverExit(&g_dfDrvObj);
    dfVxwSemPoolUninit();
    dfVxwLogUninit();
    return err;
  }
  g_bPciDriverRegistered = TRUE;

  return err;
}

void
dfVxbExit(
  void)
{
  dfDebugPrint(1, ("dfVxwExit: entered\n"));

  if (g_bPciDriverRegistered) {
    dfDebugPrint(1, ("dfVxwExit: unregistering vxBus driver\n"));
    vxbDriverUnregister(&g_dfVxbDevRegistration.b);
    g_bPciDriverRegistered = FALSE;
  }

  if (NULL != g_pDfVxbPciIdList) {
    hwMemFree((char*)g_pDfVxbPciIdList);
    g_pDfVxbPciIdList = NULL;
  }
  if (NULL != g_pPciCompatibleIdTable) {
    hwMemFree((char*)g_pPciCompatibleIdTable);
    g_pPciCompatibleIdTable = NULL;
  }

  dfVxwSemPoolUninit();
  dfVxwLogUninit();
}

#else

STATUS
dfVxbEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  dfAssert(FALSE);
  return ERROR;
}

void
dfVxbExit(
  void)
{
  dfAssert(FALSE);
}

#endif
