/*
** File: module.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for I/O module API functions.
**
** (C) Copyright Alpha Data 2010
**
** The format of the FRU ROM used in non-legacy XRMs is documented in ad-an-0001.
*/

#include <df.h>
#include "device.h"
#include "module.h"
#include <ioctl_admxrc3.h>
#include <admxrc3/types.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))

static void
parseFruRom(
  IOCTLSUBS_ADMXRC3_GETMODULEINFO* pInfo,
  uint8_t* pImage,
  unsigned int length,
  boolean_t bIsFmc,
  boolean_t* bParseError,
  boolean_t* bChecksumError)
{
  unsigned int n, index = 0;
  unsigned int boardAreaOffset, boardAreaLength, boardAreaLimit;
  unsigned int multiAreaOffset;
  unsigned int manufacturerLength, manufacturerLimit;
  unsigned int productLength, productLimit;
  unsigned int partNumberLength, partNumberLimit;
  unsigned int serialNumberLength, serialNumberLimit;
  unsigned int recordNumber = 0, recordLength, recordLimit, recordOffset;
  uint32_t manufactureDate;
  uint32_t voltage;
  uint8_t version, output, outputNumber;
  uint8_t recordId, recordFormat;
  uint8_t languageCode;
  uint8_t checksum;
  boolean_t bFoundXrmSupply = FALSE;
  boolean_t bCommonAreaOk = TRUE, bBoardAreaOk = TRUE, bMultiRecordAreaOk = TRUE;
  boolean_t bChecksumOk = TRUE;
  dfAssert(pImage != NULL);

  /*
  ** ----------------------------------------------------
  ** Parse common header area
  ** ----------------------------------------------------
  */

  /* Check version, get board area offset and get multi record offset */
  if (index >= length) {
    bCommonAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto done;
  }
  version = pImage[index];
  dfDebugPrint(4, ("parseFruRom: (Common area) version=%u(0x%02x)\n", (unsigned int)version, (unsigned int)version));
  if (version != 0x1U) {
    bCommonAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected common header format version, version=0x%x\n", version));
    goto done;
  }
  index += 3;

  /* Get board area offset */
  if (index >= length) {
    bCommonAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto done;
  }
  boardAreaOffset = pImage[index] << 3;
  dfDebugPrint(4, ("parseFruRom: (Common area) boardAreaOffset=%u(0x%02x)\n", (unsigned int)boardAreaOffset, (unsigned int)boardAreaOffset));
  index += 2;

  /* Get multi record area offset */
  if (index >= length) {
    bCommonAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto done;
  }
  multiAreaOffset = pImage[index] << 3;
  dfDebugPrint(4, ("parseFruRom: (Common area) multiAreaOffset=%u(0x%02x)\n", (unsigned int)multiAreaOffset, (unsigned int)multiAreaOffset));
  index += 2;

  /* Get common area checksum */
  if (index >= length) {
    bCommonAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto done;
  }
  checksum = pImage[index];
  dfDebugPrint(4, ("parseFruRom: (Common area) checksum=%u(0x%02x)\n", (unsigned int)checksum, (unsigned int)checksum));

  /* TO DO - verify common area checksum */

  /*
  ** ----------------------------------------------------
  ** Parse board area
  ** ----------------------------------------------------
  */

  index = boardAreaOffset;

  /* Get board area version */
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  version = pImage[index];
  dfDebugPrint(4, ("parseFruRom: (Board area) version=%u(0x%02x)\n", (unsigned int)version, (unsigned int)version));
  if (version != 0x1U) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected board area format version, version=0x%02x\n", (unsigned int)version));
    goto boardAreaDone;
  }
  index++;

  /* Get length of board area */
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  boardAreaLength = pImage[index] << 3;
  boardAreaLimit = boardAreaOffset + boardAreaLength;
  dfDebugPrint(4, ("parseFruRom: (Board area) boardAreaLength=0x%x\n", (unsigned int)boardAreaLength));
  if (boardAreaLimit > length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: board area extends past end of image, boardAreaLimit=0x%x length=0x%x\n", boardAreaLimit, length));
  }
  index++;

  /* Get language code */
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  languageCode = pImage[index];
  dfDebugPrint(4, ("parseFruRom: (Board area) languageCode=%u(0x%x)\n", (unsigned int)languageCode, (unsigned int)languageCode));
  index++;

  /* Get manufacture date as number of minutes since 00:00 01/01/1996 */
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  manufactureDate = pImage[index];
  index++;
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  manufactureDate |= ((uint32_t)pImage[index]) << 8;
  index++;
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  manufactureDate |= ((uint32_t)pImage[index]) << 16;
  index++;
  pInfo->manufactureDate = manufactureDate;
  dfDebugPrint(4, ("parseFruRom: (Board area) manufactureDate=%u(0x%06x)\n", (unsigned int)manufactureDate, (unsigned int)manufactureDate));

  /* Get manufacturer string */
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  manufacturerLength = pImage[index] & 0x3FU;
  index++;
  manufacturerLimit = index + manufacturerLength;
  if (manufacturerLimit > boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: manufacturer string extends past end of board area, manufacturerLimit=0x%x boardAreaLimit=0x%x\n", manufacturerLimit, boardAreaLimit));
  }
  if (manufacturerLimit >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: manufacturer string extends past of image, index=0x%x length=0x%x\n", manufacturerLimit, length));
    goto boardAreaDone;
  }
  n = (unsigned int)MIN(sizeof(pInfo->manufacturer), manufacturerLength);
  dfCopyKK(pInfo->manufacturer, &pImage[index], n);
  pInfo->manufacturer[sizeof(pInfo->manufacturer) - 1] = '\0';
  index += manufacturerLength;

  /* Get product name */
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  productLength = pImage[index] & 0x3FU;
  index++;
  productLimit = index + productLength;
  if (productLimit > boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: product string extends past end of board area, productLimit=0x%x boardAreaLimit=0x%x\n", productLimit, boardAreaLimit));
  }
  if (productLimit >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: product string extends past end of image, productLimit=0x%x length=0x%x\n", productLimit, length));
    goto boardAreaDone;
  }
  n = (unsigned int)MIN(sizeof(pInfo->product), productLength);
  dfCopyKK(pInfo->product, &pImage[index], n);
  pInfo->product[sizeof(pInfo->product) - 1] = '\0';
  index += productLength;

  /* Get serial number */
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  serialNumberLength = pImage[index] & 0x3FU;
  index++;
  serialNumberLimit = index + serialNumberLength;
  if (serialNumberLimit > boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: serial number string extends past end of board area, serialNumberLimit=0x%x boardAreaLimit=0x%x\n", serialNumberLimit, boardAreaLimit));
  }
  if (serialNumberLimit >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: serial number string extends past end of image, serialNumberLimit=0x%x length=0x%x\n", serialNumberLimit, length));
    goto boardAreaDone;
  }
  n = (unsigned int)MIN(sizeof(pInfo->serialNumber), serialNumberLength);
  dfCopyKK(pInfo->serialNumber, &pImage[index], n);
  pInfo->serialNumber[sizeof(pInfo->serialNumber) - 1] = '\0';
  index += serialNumberLength;

  /* Get part number */
  if (index >= boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of board area reached, index=0x%x boardAreaLimit=0x%x\n", index, boardAreaLimit));
  }
  if (index >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
    goto boardAreaDone;
  }
  partNumberLength = pImage[index] & 0x3FU;
  index++;
  partNumberLimit = index + partNumberLength;
  if (partNumberLimit > boardAreaLimit) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: part number string extends past end of board area, partNumberLimit=0x%x boardAreaLimit=0x%x\n", partNumberLimit, boardAreaLimit));
  }
  if (partNumberLimit >= length) {
    bBoardAreaOk = FALSE;
    dfDebugPrint(3, ("*** parseFruRom: part number string extends past end of image, partNumberLimit=0x%x length=0x%x\n", partNumberLimit, length));
    goto boardAreaDone;
  }
  n = (unsigned int)MIN(sizeof(pInfo->partNumber), partNumberLength);
  dfCopyKK(pInfo->partNumber, &pImage[index], n);
  pInfo->partNumber[sizeof(pInfo->partNumber) - 1] = '\0';
  index += partNumberLength;

boardAreaDone:
  /*
  ** ----------------------------------------------------
  ** Parse multi record area
  ** ----------------------------------------------------
  */

  index = multiAreaOffset;
  recordNumber = 0;

  while (1) {
    recordOffset = index;

    /* Get record ID */
    if (index >= length) {
      bMultiRecordAreaOk = FALSE;
      dfDebugPrint(3, ("*** parseFruRom: record %u begins past end of image, recordOffset=0x%x length=0x%x\n", recordNumber, recordOffset, length));
      goto multiRecordAreaDone;
    }
    recordId = pImage[index];
    index++;

    /* Get record format */
    if (index >= length) {
      bMultiRecordAreaOk = FALSE;
      dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
      goto multiRecordAreaDone;
    }
    recordFormat = pImage[index];
    index++;

    /* Get record length */
    if (index >= length) {
      bMultiRecordAreaOk = FALSE;
      dfDebugPrint(3, ("*** parseFruRom: unexpected end of image reached, index=0x%x length=0x%x\n", index, length));
      goto multiRecordAreaDone;
    }
    recordLength = pImage[index];
    recordLimit = recordOffset + recordLength + 5U;
    if (recordLimit > length) {
      bMultiRecordAreaOk = FALSE;
      dfDebugPrint(3, ("*** parseFruRom: record %u extends past end of image, recordLimit=0x%x boardAreaLimit=0x%x\n", recordNumber, recordLimit, length));
      goto multiRecordAreaDone;
    }
    index += 3;

    dfDebugPrint(3, ("parseFruRom: record %u offset=0x%x length=0x%x id=0x%x format=0x%x\n",
      recordNumber, recordOffset, recordId, recordFormat));

    if ((recordFormat & 0xfU) == 0x2U) {
      /* Format 0x2 */
      switch (recordId) {
      case 0x02U:
        /* DC load record */
        output = pImage[index];
        if (bIsFmc) {
          outputNumber = 0x0U; /* FMC uses ouput 0 (according to FRU ROM) for XRM VI/O supply */
        } else {
          outputNumber = 0x1U; /* XRM uses ouput 1 (according to FRU ROM) for XRM VI/O supply */
        }
        if ((output & 0xFU) == outputNumber) {
          /* This is the XRM VI/O supply */
          voltage = (uint32_t)pImage[index + 1] | ((uint32_t)pImage[index + 2] << 8); /* In 10 mV increments */
          /* Convert to a fixed point 16:16 number representing voltage */
          voltage <<= 16;
          voltage = voltage / 100;
          pInfo->ioVoltage.intPart = (int16_t)(voltage >> 16);
          pInfo->ioVoltage.fracPart = (uint16_t)voltage;
          bFoundXrmSupply = TRUE;
        } else {
          /* Not the XRM VI/O supply */
        }
        break;

      default:
        /* Unknown record - ignore it */
        break;
      }
    } else {
      /* Don't understand this format */
    }

    /* Move to next record */
    index = recordLimit;
    recordNumber++;

    if (recordFormat & 0x80U) {
      /* Last record */
      break;
    }
  }

multiRecordAreaDone:

  if (!bFoundXrmSupply) {
    dfDebugPrint(3, ("*** parseFruRom: did not find XRM VI/O supply\n"));
  }

done:
  /* Succeeded in parsing everything we are interested in */
  *bParseError = (bCommonAreaOk && bBoardAreaOk && bMultiRecordAreaOk && bFoundXrmSupply) ? FALSE : TRUE;
  *bChecksumError = bChecksumOk ? FALSE : TRUE;
}

DfIoStatus
ioctlGetModuleInfo(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS_ADMXRC3_GETMODULEINFO* pIoctl = (IOCTLS_ADMXRC3_GETMODULEINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  uint32_t moduleIndex;
  CoreIoModuleStatus moduleStatus;
  CoreIoModuleInfo moduleInfo;
  IOCTLSUBS_ADMXRC3_GETMODULEINFO stagingBuf;
  IOCTLSUBS_ADMXRC3_GETMODULEINFO* pUserBuf;
  boolean_t bParseError, bChecksumError;

  moduleIndex = pIoctl->in.moduleIndex;
  pUserBuf = pIoctl->in.pBuffer;
  moduleStatus = pDevCtx->coreInterface.pGetIoModuleInfo(pDevCtx->pCoreContext, moduleIndex, &moduleInfo);
  if (CoreIoModuleSuccess != moduleStatus) {
    switch (moduleStatus) {
    case CoreIoModuleInvalidIndex:
      return DfIoStatusError(ADMXRC3_INVALID_INDEX);

    default:
      return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
    }
  }

  dfZeroMemory(&stagingBuf, sizeof(stagingBuf));
  if (moduleInfo.bPresent) {
    stagingBuf.bPresent = TRUE;
    stagingBuf.flags = 0;
    if (moduleInfo.bLegacyXrm) {
      stagingBuf.flags |= ADMXRC3_MODULE_LEGACYXRM;
    }
    if (moduleInfo.bForce2v5) {
      stagingBuf.flags |= ADMXRC3_MODULE_FORCE2V5;
    }
    if (!moduleInfo.bLegacyXrm) {
      if (NULL == moduleInfo.pVpdBuffer) {
        stagingBuf.flags |= ADMXRC3_MODULE_ROMERROR;
      } else {
        parseFruRom(&stagingBuf, moduleInfo.pVpdBuffer, moduleInfo.vpdLength, moduleInfo.bIsFmc, &bParseError, &bChecksumError);
        if (bParseError) {
          stagingBuf.flags |= ADMXRC3_MODULE_ROMERROR;
        }
        if (bChecksumError) {
          stagingBuf.flags |= ADMXRC3_MODULE_CSUMERROR;
        }
      }
    }
  } else {
    stagingBuf.bPresent = FALSE;
  }

  DF_USER_SPACE_BEGIN(TAG1)

  dfUserBufferValidate(pUserBuf, sizeof(*pUserBuf), TRUE, TAG1);

  dfCopyKU(pUserBuf, &stagingBuf, sizeof(*pUserBuf), TAG1);

  DF_USER_SPACE_EXCEPT(TAG1)

  status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);

  DF_USER_SPACE_END(TAG1)

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlGetModuleInfoThunk(
	Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_GETMODULEINFO* pIoctl32 = (IOCTLS32_ADMXRC3_GETMODULEINFO*)pBuffer;
  IOCTLS_ADMXRC3_GETMODULEINFO ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.moduleIndex = pIoctl32->in.moduleIndex;
  ioctl.in.pBuffer = (IOCTLSUBS_ADMXRC3_GETMODULEINFO*)dfThunkPtr(pIoctl32->in.pBuffer);
  return ioctlGetModuleInfo(pDevCtx, &ioctl, sizeof(ioctl.in));
}
#endif
